import os
import copy
import appdirs

from . import io

try:
  import yaml
except:
  pass

_DEFAULT_CFG = dict(
  debug = False,
  max_directory_depth = 5,
  graphics_extensions = ['jpg', 'jpeg', 'pdf', 'png'],
  img_dir_name = 'img',
  img_search_paths = ['~/Documents'],
  img_search_priority = 'path-order, newest',
  input_search_paths = ['~/Documents'],
  input_search_priority = 'path-order, newest',
  bibtex_extensions = ['bib', 'bibtex'],
  bib_search_paths = ['~/Documents'],
  bib_search_priority = 'path-order, newest',
  bib_repair = dict(
    make_all_items_lowercase = True,
    merge_all_found_info = True,
    remove_fields = ['abstract'],
    generate_url_from_doi = True,
    repair_doi = True,
    verify_url_exists = True,
    verify_doi_exists = True,
    doi_and_url_verification_max_age_months = 1,
    force_double_hyphen_in_pages = True,
    keep_only_startpage_in_pages = False,
    convert_month_to_number = True,
    convert_journal_to_iso4_abbr = True,
    convert_journal_to_full_name = False,
    protect_capitalization_if_unprotected = True
  ),
  clean=dict(
    trash_folder = '~/Desktop/paperman-trash',
    ignore_paths = [],
    required_suffs = ['tex', 'pdf', 'aux'],
    clean_suffs = ['aux', 'bcf', 'log', 'snm', 'vrb', 'run.xml', 'synctex', 'nav',
                   'blg', 'fls', 'fdb_latexmk', 'toc', 'tdo', 'out', 'bak']
  ),
  lint=dict(
    avoid_commands_in_toplevel=[],
    avoid_commands = [],
    avoid_words=[],
    known_authors=[],
#    check_spelling=[]
  ),
  library_path = '~/Documents/bibliography',
  library_sync_additional_paths = [],
  library_collect_paths = ['~/Desktop', '~/Downloads'],
  library_folder_pattern = '%Y-%m',
  library_sync_device = '',
  library_sync_max_age = 365*24*60*60,
  z_bib_words_protect_capitalization = [],
  z_bib_words_dont_protect_capitalization = [],
  z_bib_journals = {},
  z_verified_urls = [],
  z_verified_dois = []
)

_REQUIRED_CFG = dict()

_CFG_PATH = os.path.join(appdirs.user_config_dir('paperman'), 'paperman.conf')
_CFG = {}
_IS_CFG_LOADED = False
_HAS_TESTED_REQUIRED_KEYS = False


def _touchCfg():
  os.makedirs(os.path.dirname(_CFG_PATH), exist_ok=True)
  if not os.path.exists(_CFG_PATH):
    with open(_CFG_PATH, 'w') as f:
      pass


def _loadCfg():
  global _IS_CFG_LOADED, _CFG, _HAS_TESTED_REQUIRED_KEYS
  if not _IS_CFG_LOADED:
    _touchCfg()
    with open(_CFG_PATH, 'r') as f:
      _CFG = yaml.safe_load(f) or {}
    _IS_CFG_LOADED = True
    _HAS_TESTED_REQUIRED_KEYS = False
    testIfRequiredExist()
    _writeCfg()


def _writeCfg():
  _touchCfg()
  with open(_CFG_PATH, 'w') as f:
    yaml.dump({k: v for k, v in {**_DEFAULT_CFG, **_CFG}.items()}, f)


def path():
  return os.path.abspath(_CFG_PATH)


def reload():
  global _IS_CFG_LOADED
  try:
    _IS_CFG_LOADED = False
    _loadCfg()
  except (KeyboardInterrupt, RuntimeError):
    raise
  except:
    io.err(io.coercedStacktrace(10))
    io.err('error reloading config file, reverting '
           'changes on disk')
    _writeCfg()
    _loadCfg()


def testIfRequiredExist(*path):
  global _HAS_TESTED_REQUIRED_KEYS, _IS_CFG_LOADED
  if not _HAS_TESTED_REQUIRED_KEYS:
    settings = _REQUIRED_CFG
    for p in path:
      settings = settings[p]
    if type(settings) is dict:
      for k in settings.keys():
        testIfRequiredExist(*path, k)
    else:
      if get(*path) is None:
        _IS_CFG_LOADED = False
        raise RuntimeError(f'config setting {".".join(path)} is required '
                           f'but is not set')
  if len(path) == 0:
    _HAS_TESTED_REQUIRED_KEYS = True


def get(*key):
  if len(key) < 1:
    raise ArgumentError('cfg.get() expects at least one argument: key')
  _loadCfg()
  setting, default = _CFG, _DEFAULT_CFG
  usedDefault = False
  for k in key:
    if k not in setting.keys():
      usedDefault = True
      setting[k] = default[k]
    default = default[k]
    setting = setting[k]

  if usedDefault:
    _writeCfg()

  return setting


def set(*keyVal):
  if len(keyVal) < 2:
    raise ArgumentError('cfg.set() expects at least two arguments: key and value')
  key = keyVal[:-1]
  val = nativify(keyVal[-1])
  reload()
  d = _CFG
  for k in key[:-1]:
    if k not in d:
      d[k] = {}
    d = d[k]
  d[key[-1]] = val
  _writeCfg()


def nativify(data, skipIfFunc=lambda data: False, skipIfError=False):
  args = dict(skipIfFunc=skipIfFunc, skipIfError=skipIfError)

  if data is None or type(data) in (bool, int, float, complex, str):
    return data

  if type(data) is dict:
    return {nativify(k, **args): nativify(v, **args) for k, v in data.items()}

  if hasattr(data, '__iter__'):
    try:
      return [nativify(v, **args) for v in data]
    except:
      pass

  try:
    return float(data)
  except:
    try:
      return complex(data)
    except:
      pass

  if skipIfError or skipIfFunc(data):
    return None

  raise ValueError('could not convert data to native type: "'+str(data)+'"')



from .common import *

def main(args):
  proj = detectProj(args)
  if proj is None:
    return

  allFine = True
  for l in proj.lint():
    #io.dbg(f'proj.lint yielded {l}')
    f, ln, msg = l
    io.info(f'in file {f}, line {ln}:', msg)
    allFine = False

  if allFine:
    io.info('all fine')



import shutil
import os
import fnmatch

from . import io
from . import cfg
from . import parser


def _candidateSortKey(rules):
  def key(c, rules=rules):
    pathNo, path = c[:2]
    res = ''
    for r in rules:
      if 'path-order' in r:
        res += f'_{pathNo:04d}'
      elif 'newest' in r:
        res += f'_{2**64-os.path.getmtime(path)}'
      elif 'oldest' in r:
        res += f'_{os.path.getmtime(path)}'
      else:
        raise RuntimeError(f'unknown priority rule {r}')
    return res
  return key


def _fastGlob(path):
  # walk through filesystem sub tree described by path up to the first '*'
  basePath = path.split('*')[0]

  # remove any inner '../' and './' from match path, as they dont make sense there
  matchPath = path.replace('/../', '/').replace('/./', '/')

  res = []
  hasWarnedDepth = False
  io.dbg(f'calling _fastGlob("{path}"), walking through "{basePath}"')
  for root, dirs, files in os.walk(basePath, topdown=True):
    # abort tree is too deep
    if root.count(os.sep)-basePath.count(os.path.sep)-2 > cfg.get('max_directory_depth'):
      io.verb(f'skipping subfolders of {root}')
      if not hasWarnedDepth:
        hasWarnedDepth = True
        io.warn(f'reached max_directory_depth='
                f'{cfg.get("max_directory_depth")} when recursing '
                f'through the current directory, ignoring deeper '
                f'levels')
      dirs.clear()

    # do not recurse into links
    while True:
      i = [d for d in dirs if os.path.islink(d)]
      if not i:
        break
      io.dbg(f'skipped link {i[0]}')
      dirs.remove(i[0])


    # skip hidden directories (e.g. git)
    while True:
      i = [d for d in dirs if d.startswith('.')]
      if not i:
        break
      #io.dbg(f'skipped hidden directory {i[0]}')
      dirs.remove(i[0])

    # check if any file matches
    for f in files:
      fname = os.path.join(root, f)
      #io.dbg(f' checked filename {fname} for match with {matchPath}')
      if fnmatch.fnmatch(fname, matchPath):
        res.append(fname)
  return res


def importImgs(imgs, imgDir):
  success, failed = [], []

  # generate search paths
  searchPaths = []
  for p in cfg.get('img_search_paths'):
    # expand ~ in path
    p = os.path.expanduser(p)

    # append unedited path
    searchPaths.append(p)

    # if no **-glob pattern detected, add standard one and add to list
    if '*' not in p:
      searchPaths.append(os.path.join(p, '**', cfg.get('img_dir_name')))

  # iterate through requested images and try to import
  for img in imgs:
    candidates = []
    for i, p in enumerate(searchPaths):
      candidates.extend([(i, c) for c in _fastGlob(
                                        os.path.join(p, img.fname+'.*'))])

    # in case no candidate was found, add img to failed list and continue
    if not candidates:
      failed.append(img)
      continue

    # generate sorting function according to config
    rules = cfg.get('img_search_priority').split()
    sortedCandidates = sorted(candidates, key=_candidateSortKey(rules))
    io.dbg(f'first ten matching candidates sorted by {rules}:', sortedCandidates[:10])

    # if there is a pdf under the first five candidates, prefer the pdf
    for _, c in sortedCandidates[:5]:
      if c.lower().endswith('.pdf'):
        bestMatch = c
        break
    else:
      _, bestMatch = sortedCandidates[0]

    # copy best match to desired location
    os.makedirs(imgDir, exist_ok=True)
    shutil.copy(bestMatch, imgDir)
    io.verb(f'copying {bestMatch} -> {imgDir}')
    success.append(img)

  return success, failed


def importCites(cites):
  success, failed = [], []

  # generate search paths
  searchPaths = []
  for p in (list(cfg.get('bib_search_paths'))
            + [cfg.get('library_path')+'/**'] if cfg.get('library_path') else []):
    # expand ~ in path
    p = os.path.expanduser(p)

    # append unedited path
    searchPaths.append(p)

  # create huge list with all found bib paths
  allCites = []
  for i, p in enumerate(searchPaths):
    for ext in cfg.get('bibtex_extensions'):
      for f in _fastGlob(os.path.join(p, '*.'+ext)):
        try:
          _cites = parser.BibFile(f).cites()
        except RuntimeError as e:
          io.warn('error parsing bibtex file on search path:',
                  str(e),
                  'skipping...')
        else:
          for c in _cites:
            if c not in [c for _, _, c in allCites]:
              allCites.append([i, f, c])

  # iterate through requested images and try to import
  for cite in cites:
    candidates = [(i, f, c) for i, f, c in allCites if c == cite]

    # in case no candidate was found, add cite to failed list and continue
    if not candidates:
      failed.append(cite)
      continue

    # generate sorting function according to config
    rules = cfg.get('bib_search_priority').split()
    sortedCandidates = sorted(candidates, key=_candidateSortKey(rules))
    io.dbg(f'first ten matching candidates sorted by {rules}:', sortedCandidates[:10])
    _, _path, bestMatch = sortedCandidates[0]
    io.verb(f'best match for citation with key "{cite.key}"',
            f'found in file',
            f'{_path}')

    if cfg.get('bib_repair', 'merge_all_found_info'):
      if len(sortedCandidates) > 1:
        io.verb('merging with info from all other candidates...')
      for c in sortedCandidates[1:]:
        bestMatch[-1].insertNonexistingItems(c[-1])

    success.append(bestMatch)

  return success, failed


def importInclude(include):
  success, failed = [], []

  if not os.path.exists(include):
    # generate search paths
    searchPaths = [os.path.expanduser(p) for p in cfg.get('input_search_paths')]

    # try to import
    candidates = []
    for i, p in enumerate(searchPaths):
       candidates.extend([(i, c) for c in _fastGlob(
                                os.path.join(p, os.path.basename(include)))])

    # in case no candidate was found, add input to failed list and return
    if not candidates:
      failed.append(include)
      return False

    # generate sorting function according to config
    rules = cfg.get('input_search_priority').split()
    sortedCandidates = sorted(candidates, key=_candidateSortKey(rules))
    io.dbg(f'first ten matching candidates sorted by {rules}:',
           sortedCandidates[:10])
    _, bestMatch = sortedCandidates[0]

    # copy best match to desired location
    os.makedirs(os.path.dirname(include) or '.', exist_ok=True)
    shutil.copy(bestMatch, include)
    io.verb(f'copying {bestMatch} -> {include}')
    return True
  return False



# PYTHON_ARGCOMPLETE_OK

import argcomplete
import argparse

from . import io
from . import cfg
from . import utils


@utils.logExceptionsAndRaise
def main():
  # show startup message
  io.startup()

  # options used by multiple subparsers
  def addVerboseArg(parser):
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='print more detailed output')
  def addTexFileArg(parser):
    parser.add_argument('tex_file', default='', nargs='?',
                        help='toplevel tex file of the document')

  # setup argparser and global arguments
  p = argparse.ArgumentParser()
  addVerboseArg(p)
  p.add_argument('--version', action='store_true',
                 help='print installed paperman version')

  sub = p.add_subparsers(metavar='subcommands', dest='command')

  # img command
  s = sub.add_parser('img', help='image helper: check if unused or missing '
                                 'images exist or import missing images')
  addTexFileArg(s)
  addVerboseArg(s)
  s.add_argument('-i', '--import', action='store_true',
                 help='try importing missing images from image search path')
  s.add_argument('-c', '--clean', action='store_true',
                 help='remove unused images')

  # bib command
  s = sub.add_parser('bib', help='bibliography helper: check if unused or '
                                 'missing bib entries exist or import '
                                 'missing entries')
  addTexFileArg(s)
  addVerboseArg(s)
  s.add_argument('-p', '--print', action='store_true',
                 help='try to find missing citations in existing bibliographies '
                      'and show them in the shell, conflicts with --import, '
                      '--rewrite, and --sort')
  s.add_argument('-i', '--import', action='store_true',
                 help='try importing missing citations from existing '
                      'bibliographies')
  s.add_argument('-c', '--clean', action='store_true',
                 help='remove unused citations from bibliography, implies '
                      '--rewrite')
  s.add_argument('-r', '--rewrite', action='store_true',
                 help='rewrite and repair entire bibliography')
  s.add_argument('-s', '--sort', nargs='?', default=False,
                 help='sort bibliography, sort order can be one of "date", '
                      '"author" or "key", default is "key", implies --rewrite')

  # input command
  s = sub.add_parser('input',
                     help=r'input helper: check if all \input{} files exist '
                          r'or import files')
  addTexFileArg(s)
  addVerboseArg(s)
  s.add_argument('-i', '--import', action='store_true',
                 help=r'try importing missing \input{} files from input '
                      r'search path')

  # call img, bib and input command with --import flag set
  s = sub.add_parser('import-all',
                     help='shortcut to run img, bib and input subcommands '
                          'with --import option enabled')
  addTexFileArg(s)
  addVerboseArg(s)

  # sort authors command
  s = sub.add_parser('sort-authors',
                     help='alphabetic sorting and formatting of author lists '
                          'for scientific publications and presentations')
  addVerboseArg(s)
  s.add_argument('authors', nargs='+',
                 help='comma separated list of authors seperated. Use -s '
                      'option for a different separator')
  s.add_argument('-s', '--separator', default=',',
                 help='separation character in author list')
  s.add_argument('-f', '--full-name', action='store_true',
                 help='output full author names.')
  s.add_argument('--keep-first', action='store_true',
                 help='keep position of first author in list.')
  s.add_argument('--keep-last', action='store_true',
                 help='keep position of last author in list.')
  s.add_argument('-k', '--keep-first-and-last', action='store_true',
                 help='keep position of first and last author in list.')
  s.add_argument('-q', '--quiet', action='store_true',
                 help='just output the sorted author list, nothing else.')
  s.add_argument('--no-and', action='store_true',
                 help='use the separator instead of "and" before the final '
                      'name and ignore ands in input.')

  # collect bib/pdf pairs from specified collect directories
  s = sub.add_parser('collect',
                     help='rename and move bib-pdf file pairs from specified '
                          'folders to library')
  addVerboseArg(s)
  s.add_argument('paths', nargs='*',
                 help='specify paths to import bib-pdf file pairs from, '
                      'defaults to paths given in config')
  s.add_argument('-w', '--watch', action='store_true',
                 help='never exit and keep watching the collect directories')
  s.add_argument('-e', '--err-to-file', action='store_true',
                 help='write errors not only to stdout, but also to a txt file '
                      'located in the same position as the bib-pdf file pair '
                      'that caused the error')

  # library subcommand
  s = sub.add_parser('lib',
                     help='search library, check library health, detect '
                          'corrupt bib files, duplicates and possibly '
                          'broken pdfs')
  addVerboseArg(s)
  s.add_argument('-f', '--find', nargs='+',
                 help='find library entries that contain given words in '
                      'bibtex fields')
  s.add_argument('-F', '--find-fulltext', nargs='+',
                 help='find library entries that contain given words in '
                      'full manuscript text (requires pdf2txt program to '
                      'to be installed on your $PATH)')
  s.add_argument('-k', '--key', action='store_true',
                 help='print citation key instead of path')
  s.add_argument('-l', '--long', action='store_true',
                 help='print full bibtex entries instead of path only')
  #s.add_argument('-L', '--extra-long', action='store_true',
  #               help='print full pdf2txt result instead of path only')

  # sync subcommand
  s = sub.add_parser('sync',
                     help='sync library pdfs to a tablet or other device')
  addVerboseArg(s)
  s.add_argument('-p', '--sync-path',
                 help='path to the device root that is to be synced.')

  # diff subcommand
  s = sub.add_parser('diff',
                     help='build pdfs that show changes between file versions')
  addVerboseArg(s)
  s.add_argument('-t' , '--old-is-tag', action='store_true',
                 help='do not treat "old" argument as filename, '
                      'but as git tag name (requires git to be installed '
                      'in your $PATH, requires current directory to be '
                      'git repository)')
  s.add_argument('-T' , '--new-is-tag', action='store_true',
                 help='do not treat "new" argument as filename, '
                      'but as git tag name (requires git to be installed '
                      'in your $PATH, requires current directory to be '
                      'git repository)')
  s.add_argument('-o', '--outfile', nargs='?', default='diff.pdf',
                 help='name of the pdf file to be built')
  s.add_argument('-k', '--keep', action='store_true',
                 help='keep build files build files')
  s.add_argument('-c', '--clean', action='store_true',
                 help='remove all build files except for output pdf')
  s.add_argument('old', help='filename of old version')
  s.add_argument('new', help='filename of new version')
  s.add_argument('filename', nargs='?',
                 help='only used if both --old-is-tag and '
                      '--new-is-tag are set, specifies filename')

  # make-diff subcommand
  s = sub.add_parser('journal',
                     help='find full journal names and abbreviations')
  addVerboseArg(s)
  s.add_argument('-f', '--full', action='store_true',
                 help='print full names of found journals')
  s.add_argument('-a', '--abbreviated', action='store_true',
                 help='print abbreviated names of found journals ')
  s.add_argument('--list-suspicious', action='store_true',
                 help='print all journals in database that might by errornous, '
                      'i.e., abbreviations that do not end with a dot')
  s.add_argument('name', nargs='*',
                 help='journal name, abbreviation or partial name')

  # lint subcommand
  s = sub.add_parser('lint',
                     help='search latex project for potential errors')
  addVerboseArg(s)
  addTexFileArg(s)

  # clean subcommand
  s = sub.add_parser('clean',
                     help='recurse into subfolders of current directory and '
                          'remove latex build files')
  s.add_argument('-o', '--only-old', action='store_true',
                 help='only clean if build files have not changed since '
                      'more than one month')
  addVerboseArg(s)

  # config subcommand
  s = sub.add_parser('config',
                     help='print location of config file and exit')
  s.add_argument('-o', '--open', action='store_true',
                 help='open config file in vim')
  addVerboseArg(s)

  # enable autocompletion and parse args
  argcomplete.autocomplete(p)
  args = p.parse_args()

  # apply verbosity setting
  io.isVerbose = args.verbose

  # if version is requested, print version and exit
  if args.version:
    io.info(f'paperman version {io.__version__}')
    return

  # select submodule for subcommands
  cmd, cmds = None, None
  if not args.command:
    io.err('subcommand is required')
    return

  elif args.command == 'img':
    from .subcommands import img as cmd

  elif args.command == 'bib':
    from .subcommands import bib as cmd

  elif args.command == 'input':
    from .subcommands import inp as cmd

  elif args.command == 'import-all':
    from .subcommands import inp as cmd1
    from .subcommands import img as cmd2
    from .subcommands import bib as cmd3
    setattr(args, 'import', True)
    args.clean = False
    args.rewrite = False
    args.sort = False
    args.print = False
    cmds = [cmd1, cmd2, cmd3]

  elif args.command == 'sort-authors':
    from .subcommands import sort_authors as cmd

  elif args.command == 'collect':
    from .subcommands import collect as cmd

  elif args.command == 'lib':
    from .subcommands import lib as cmd

  elif args.command == 'sync':
    from .subcommands import sync as cmd

  elif args.command == 'diff':
    from .subcommands import diff as cmd

  elif args.command == 'journal':
    from .subcommands import journal as cmd

  elif args.command == 'lint':
    from .subcommands import lint as cmd

  elif args.command == 'clean':
    from .subcommands import clean as cmd

  elif args.command == 'config':
    from .subcommands import config as cmd

  else:
    io.dbg(f'args={args}')
    raise ValueError(f'args.command has unexpected value')


  # raise error if required config keys are missing
  cfg.testIfRequiredExist()

  # run subcommand module
  cmdStr = str(cmd or '\n'+'\n'.join([str(c) for c in cmds]))
  io.verb(f'selected subcommand(s): {cmdStr}')
  try:
    if cmd:
      cmds = [cmd]
    if cmds:
      for cmd in cmds:
        cmd.main(args)
  except KeyboardInterrupt:
    raise
  except RuntimeError as e:
    io.err(str(e))
  except Exception as e:
    if cfg.get('debug'):
      raise
    io.err(str(e))


if __name__ == '__main__':
  main()



import shutil
import subprocess

from .. import utils
from .common import *


def main(args):
  buildDir = '.'
  if args.old_is_tag and args.new_is_tag:
    buildDir = os.path.dirname(args.filename)
  elif args.old_is_tag:
    buildDir = os.path.dirname(args.new)
  elif args.new_is_tag:
    buildDir = os.path.dirname(args.old)
  else:
    buildDir = os.path.dirname(args.new)
  buildDir = buildDir or '.'

  # if clean option is set, just clean and return
  if args.clean:
    for f in os.listdir(buildDir):
      if f.startswith('paperman-diff'):
        fname = os.path.join(buildDir, f)
        os.remove(fname)
        io.verb(f'removing {fname}')
    return

  # check if build files exist and warn about
  # overwriting them
  if any([f.startswith('paperman-diff') for f in os.listdir(buildDir)]):
    if not io.conf(f'filenames beginning with "paperman-diff" exist',
                   f'in "{(buildDir or ".")+os.path.sep}" and may be overwritten',
                   f'during diff build, continue?', default=True):
      return

  # function to copy file versions
  def copyToBuildDir(path, name, buildDir=buildDir):
    texSrc = utils.replaceSuffix(path, 'tex')
    texTar = os.path.join(buildDir, name+'.tex')
    shutil.copy(texSrc, texTar)
    io.verb(f'copied {texSrc} -> {texTar}')

    bblSrc = utils.replaceSuffix(path, 'bbl')
    bblTar = os.path.join(buildDir, name+'.bbl')
    if os.path.exists(bblSrc):
      shutil.copy(bblSrc, bblTar)
      io.verb(f'copied {bblSrc} -> {bblTar}')
    elif os.path.exists(bblTar):
      os.remove(bblTar)
      io.verb(f'removed {bblTar}')

  # function to extract tagged file versions
  def gitShowToBuildDir(tag, path, name):
    texPath = utils.replaceSuffix(path, 'tex')
    r = subprocess.run(['git', 'show', tag+':./'+texPath],
                       capture_output=True)
    if r.returncode:
      raise RuntimeError('failed to get tex file via "git show":\n'
                         +r.stdout.decode()+'\n'+r.stderr.decode())
    texTar = os.path.join(buildDir, name+'.tex')
    with open(texTar, 'w') as f:
      f.write(r.stdout.decode())
    io.verb(f'wrote result of "git show {tag}:./{texPath}" to {texTar}')

    bblPath = utils.replaceSuffix(path, 'bbl')
    bblTar = os.path.join(buildDir, name+'.bbl')
    r = subprocess.run(['git', 'show', tag+':./'+bblPath], capture_output=True)
    if r.returncode:
      io.warn('failed to get bbl file via "git show":\n'
              +r.stdout.decode()+'\n'+r.stderr.decode())
      if os.path.exists(bblTar):
        os.remove(bblTar)
        io.verb(f'removed {bblTar}')
    else:
      bblTar = os.path.join(buildDir, name+'.bbl')
      with open(bblTar, 'w') as f:
        f.write(r.stdout.decode())
      io.verb(f'wrote result of "git show {tag}:./{bblPath}" to {bblTar}')

  # prepare old files
  if args.old_is_tag:
    if args.new_is_tag:
      path = args.filename
    else:
      path = args.new
    gitShowToBuildDir(args.old, path, 'paperman-diff-old')
  else:
    copyToBuildDir(args.old, 'paperman-diff-old')

  # prepare new files
  if args.new_is_tag:
    if args.old_is_tag:
      path = args.filename
    else:
      path = args.old
    gitShowToBuildDir(args.new, path, 'paperman-diff-new')
  else:
    copyToBuildDir(args.new, 'paperman-diff-new')

  # run command and raise on error
  def run(*cmd, stdout=None, raiseOnErr=True):
    io.verb('running command: '+' '.join(cmd))
    r = subprocess.run(cmd, cwd=buildDir,
                       stderr=subprocess.PIPE,
                       stdout=stdout or subprocess.PIPE)
    if r.returncode:
      io.verb(f'command failed: {r.returncode}')
      if raiseOnErr:
        raise RuntimeError(f'{cmd[0]} failed unexpectedly\n'
                          +r.stdout.decode()+'\n'+r.stderr.decode())
      else:
        return False
    io.verb(f'command exited successful')
    return r

  # run build process
  target = os.path.join(buildDir, 'paperman-diff.pdf')
  if os.path.exists(target):
    os.remove(target)
    io.verb(f'removed {target}')

  for fs, msg in ([True, 'trying to build with -t FLOATSAFE...'],
                  [False, 'trying to build without -t FLOATSAFE...']):
  #                [None, 'trying to build without bib-diff...']):
    io.info(msg)
    for e in ('tex', 'bbl'):
      bblOld = os.path.join(buildDir, 'paperman-diff-old.bbl')
      bblNew = os.path.join(buildDir, 'paperman-diff-new.bbl')
      if (e == 'bbl'
            and (not os.path.exists(bblOld)
               or not os.path.exists(bblNew)
               or fs is None)):
        bblTar = os.path.join(buildDir, 'paperman-diff.bbl')
        if os.path.exists(bblNew):
          shutil.copy(bblNew, bblTar)
          io.verb(f'copied {bblNew} -> {bblTar}')

        elif os.path.exists(bblOld):
          shutil.copy(bblOld, bblTar)
          io.verb(f'copied {bblOld} -> {bblTar}')

        continue

      with open(os.path.join(buildDir, 'paperman-diff.'+e), 'w') as f:
        run('latexdiff', *(['-t', 'FLOATSAFE'] if e=='tex' and fs else []),
            'paperman-diff-old.'+e, 'paperman-diff-new.'+e, stdout=f)

    for _ in range(2):
      res = run('pdflatex', '-interaction=nonstopmode', 'paperman-diff',
                raiseOnErr=False) #not fs) #fs is None)

    # if successful, break
    if res:
      break

  if os.path.exists(target):
    shutil.move(os.path.join(buildDir, 'paperman-diff.pdf'), args.outfile)

    # remove build files if 'keep' flag is not set
    if not args.keep:
      for f in os.listdir(buildDir):
        if f.startswith('paperman-diff'):
          fname = os.path.join(buildDir, f)
          os.remove(fname)
          io.verb(f'removed {fname}')

    io.info(f'successfully generated {args.outfile}')
  else:
    io.err('failed to build diff')



from .common import *

p = lambda d: os.path.realpath(os.path.expanduser(d))
trashFolder = p(cfg.get('clean', 'trash_folder'))
baseDirectory = p('.')

def stripBaseDir(p):
  if p.startswith(baseDirectory):
    return p[len(baseDirectory):]
  return p


def main(args):
  proj = detectProj(args)
  if proj is None:
    return

  unused = proj.unusedIncludedImgs()
  if unused:
    io.info('detected unused image files:', *sorted(unused))
  else:
    io.info('no unused image files')

  if unused and args.clean:
    if io.conf('move all unused images to paperman trash?', default=True):
      for img in unused:
        absPath, path = img.path, stripBaseDir(img.path)
        baseFolder = time.strftime('%Y-%m-%d')
        absMoveTo = os.path.realpath(f'{trashFolder}/{baseFolder}/{path}')
        io.info(f'{stripBaseDir(absPath)} -> {stripBaseDir(absMoveTo)}')
        os.makedirs(os.path.dirname(absMoveTo), exist_ok=True)
        shutil.move(absPath, absMoveTo)
    io.info('all done.')

  missing = proj.missingIncludedImgs()
  if missing:
    io.info('detected missing images:', *sorted(missing))
  elif getattr(args, 'import'):
    io.info('no missing images, nothing to import')
  else:
    io.info('no missing images')

  if missing and getattr(args, 'import'):
    # check if image search is configured
    if len(cfg.get('img_search_paths')) == 0:
      io.err('no "img_search_paths" configured to import images from')
      return

    # if verbose, print all detected graphicspaths of toplevel tex files
    for t in proj.toplevel():
      io.verb(f'detected graphicspaths of toplevel tex file',
              f'{t.path}:',
              *[f'  {prettyDirectory(p)}' for p in t.graphicspath()])

    # print error and exit in case no iamge directory was detected
    if len(proj.commonImageDirs()) == 0:
      io.err(r'failed to detect image directory for current project,',
             r'a valid images directory has to be set with \graphicspath{}',
             r'in all toplevel tex documents of the project')
      return

    # if exactly one image directory was detected, use this directory
    elif len(proj.commonImageDirs()) == 1:
      targetDir = proj.commonImageDirs()[0]
      io.verb('found exactly one common graphicspath among all toplevel tex files:',
              targetDir)

    # if more than one image directory was detected, ask user to select one
    else:
      allDirs = proj.commonImageDirs()
      mostProbable = sorted([d for d in allDirs if isGoodImageDirectory(d)])
      remaining = [d for d in allDirs if d not in mostProbable]
      if len(mostProbable) == 1:
        default = 1
      else:
        default = None

      # ask user to select directory
      _, targetDir = io.select('found multiple image directory candidates for current '
                               'project, please select one:',
                               [prettyDirectory(o) for o in mostProbable+remaining],
                               default=default)

    # ask user to confirm if selected directory does not look like an img directory
    if not isGoodImageDirectory(targetDir):
      if not io.conf(f'the selected target folder\n"{prettyDirectory(targetDir)}"',
                     f'does not look like a good image\nfolder, a good image'
                     f'folder should be named',
                     f'"{cfg.get("img_dir_name")+os.path.sep}"',
                     f'use this target folder anyway? ',
                     default=False):
        return

    # import finder module and import all missing imgs
    from .. import finder
    success, failed = finder.importImgs(missing, targetDir)

    # report results
    if success:
      if len(success) == 1:
        io.info(f'successfully found and imported image')
      else:
        io.info(f'successfully found and imported {len(success)} images')
    if failed:
      io.info(f'could not find the following images:',
              *[i.fname for i in failed])



from .common import *

def main(args):
  proj = detectProj(args)
  if proj is None:
    return

  unused = proj.unusedCites()
  if unused:
    io.info('detected unused citations:', *sorted(unused))
  else:
    io.info('no unused citations')

  missing = proj.missingCites()
  if missing:
    io.info('detected missing citations:', *sorted(missing))
  elif getattr(args, 'import'):
    io.info('no missing citations, nothing to import')
  else:
    io.info('no missing citations')

  # prepare flags wether to import missing, sort and rewrite
  importMissing = False
  if getattr(args, 'import'):
    importMissing = True
  rewrite = args.rewrite or args.sort or args.clean
  sortKey = None
  if not args.sort:
    pass
  elif args.sort == 'key' or args.sort is None:
    sortKey = lambda c: c.key
  elif args.sort == 'date':
    sortKey = lambda c: (c['year'] or '0000')+(c['month'] or '00')
  elif args.sort == 'author':
    sortKey = lambda c: c['author'] or ''
  else:
    raise RuntimeError(f'unknown sort order "{args.sort}"')

  # check if print is present and raise error if combined with conflicting options
  showOnly = False
  if hasattr(args, 'print') and getattr(args, 'print'):
    if importMissing or rewrite or sortKey:
      raise RuntimeError('--print option cannot be combined with --import, '
                         '--sort or --rewrite options')
    showOnly = True

  if missing and (importMissing or showOnly):
    # check if bibliography search is configured
    if len(cfg.get('bib_search_paths')) == 0:
      io.err('no "bib_search_paths" configured to import citations from')
      return

    # if verbose, print all detected graphicspaths of toplevel tex files
    for t in proj.toplevel():
      if t.bibs():
        io.verb(f'detected bibliographies of toplevel tex file',
                f'{t.path}: '
                +', '.join([b.path or b.fname for b in t.bibs()]))

      # find missing cites for current toplevel file
      _missing = t.missingCites()
      if _missing:

        # import finder module and import all missing citations
        from .. import finder
        success, failed = finder.importCites(_missing)

        # report results
        if success:
          if len(success) == 1:
            io.info(f'successfully found citation')
          else:
            io.info(f'successfully found {len(success)} citations')
        if failed:
          io.info(f'could not find the following citations:',
                  *[f'"{c.key}"' for c in failed])

        existingBibs = [b for b in t.bibs() if b.exists()]
        if showOnly:
          prettyBib = '\n\n'.join(['']+[c.pretty() for c in success]+[''])
          if len(existingBibs) == 0:
            io.info(f'copy the following to the bib file of toplevel file',
                    f'"{t.path}"')
          elif len(existingBibs) == 1:
            io.info(f'copy the following to the bib file:',
                    f'"{existingBibs[0].path}"',
                    f'of toplevel file "{t.path}"')
          elif len(existingBibs) > 1:
            io.info(f'copy the following to one of the bib files:',
                    '\n'.join([f'  "{b.path}"' for b in existingBibs]),
                    f'of toplevel file "{t.path}"')
          io.raw('-'*60+prettyBib+'-'*60)
        else:
          if len(existingBibs) == 0 and len(t.bibs()) == 1:
            t.bibs()[0].create()
            existingBibs = t.bibs()

          if len(existingBibs) == 0:
            raise RuntimeError(f'failed to detect bibliography file for '
                               f'tex file "{t.path}"')
          elif len(existingBibs) == 1:
            bib = existingBibs[0]

          else:
            i, _ = io.select(f'found multiple bibliography files for tex file',
                               f'"{t.path}"',
                               f'please select one',
                               options=[b.path for b in existingBibs])
            bib = existingBibs[i]

          io.info(f'appending found citations to "{bib.path}"')
          bib.addCites(success)

  # go trough all toplevel files and apply actions
  if rewrite and not importMissing:
    io.info('rewriting bib file(s)...')
    for t in proj.toplevel():
      for bib in t.bibs():
        
        # create backup file of existing bib
        uniqueName = bib.path+'.bak'
        i = 0
        while os.path.exists(uniqueName):
          i += 1
          uniqueName = bib.path+f'.{i}.bak'
        shutil.copy(bib.path, uniqueName)

        # create list of all existing cites to manipulate on
        newCites = bib.cites()

        # remove unused citations if requested
        if unused and args.clean:
          if io.conf(f'remove all unused entries from bib file {bib.path}?', default=True):
            newCites = [c for c in newCites if c not in unused]

        # apply sorting scheme if requested
        if sortKey:
          newCites = sorted(newCites, key=sortKey)

        # write new list of citations to disk
        bib.setCites(newCites)

        # offer to delete backup bib
        if io.conf(f'delete copy of old bib file {uniqueName}?', default=False):
          os.remove(uniqueName)



#!/usr/bin/python3

from setuptools import setup
from setuptools.command.install import install

# DO NOT CHANGE: this line will be replaced by ./dev/update-setup.py
version = '1.0.9'

# if setup is run in project dir, update version number
try:
  import os
  import subprocess
  p = os.path.join(os.path.dirname(__file__), 'dev/update-setup.sh')
  if (os.path.isfile(p) and os.access(p, os.X_OK)):
    version = subprocess.run([p], capture_output=True, check=True).stdout.decode().strip()
except KeyboardInterrupt:
  raise
except:
  pass

# prepare post-install command
class PostInstall(install):
  def run(self):
    install.run(self)
    import subprocess
    subprocess.run(['activate-global-python-argcomplete']) #, check=True)

# read the contents of your README file
from os import path
with open(path.join(path.dirname(__file__), 'README.md'),
          encoding='utf-8') as f:
  description = f.read()

# run setup
setup(name='paperman',
      description='latex project and bibliography management utility',
      python_requires='>=3.6',
      long_description=description,
      long_description_content_type='text/markdown',
      author='zaphB',
      version=version,
      packages=['paperman', 'paperman.subcommands', 'paperman.parser'],
      entry_points={
        'console_scripts': [
          'paperman = paperman.__main__:main'
        ],
        'gui_scripts': []
      },
      install_requires=['appdirs', 'argparse', 'argcomplete', 'pyyaml', 'unidecode'],
      cmdclass={
        'install': PostInstall
      }
)



from .common import *

class IncludeMissing(RuntimeError):
  pass

def main(args):
  enableImport = hasattr(args, 'import') and getattr(args, 'import')

  # check if image search is configured
  if enableImport:
    if len(cfg.get('input_search_paths')) == 0:
      io.err(r'no "input_search_paths" configured to import \input{} files from')
      return

  try:
    proj = detectProj(args,
                      enableIncludeImport=enableImport,
                      raiseOnIncludeNotFound=IncludeMissing)
    if proj is None:
      return
    for t in proj.toplevel():
      t.recurseThroughIncludes()

  except IncludeMissing as e:
    if enableImport:
      io.err(r'failed to import \input{} file:',
             str(e))
    else:
      io.info(r'imported missing \input{} file:',
              str(e))

  else:
    io.info(r'all files imported by \input{} exist')
